import os
import numpy as np
from importRosbag.importRosbag import importRosbag
from .dataset import Dataset
from .download_utils import check_integrity, download_url


class VPR(Dataset):
    """Event-Based Visual Place Recognition With Ensembles of Temporal Windows <https://zenodo.org/record/4302805>. 
    Events have (txyp) ordering.

    Parameters:
        save_to (string): Location to save files to on disk.
        download (bool): Choose to download data or verify existing files. If True and a file with the same 
                    name and correct hash is already in the directory, download is automatically skipped.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        
    Returns:
        A dataset object that can be indexed or iterated over. One sample returns a tuple of (events, imu, images).
    """

    base_url = "https://zenodo.org/record/4302805/files/"
    recordings = [  # recording names and their md5 hash
        ["dvs_vpr_2020-04-21-17-03-03.bag", "04473f623aec6bda3d7eadfecfc1b2ce"],
        ["dvs_vpr_2020-04-22-17-24-21.bag", "ca6db080a4054196fe65825bce3db351"],
        ["dvs_vpr_2020-04-24-15-12-03.bag", "909569732e323ff04c94379a787f2a69"],
        ["dvs_vpr_2020-04-27-18-13-29.bag", "e80b6c0434690908d855445792d4de3b"],
        ["dvs_vpr_2020-04-28-09-14-11.bag", "7854ede61c0947adb0f072a041dc3bad"],
        ["dvs_vpr_2020-04-29-06-20-23.bag", "d7ccfeb6539f1e7b077ab4fe6f45193c"],
    ]
    
    sensor_size = (260, 346)
    ordering = "txyp"

    def __init__(self, save_to, download=True, transform=None, target_transform=None):
        super(VPR, self).__init__(
            save_to, transform=transform, target_transform=target_transform
        )
        folder_name = "visual_place_recognition"
        self.location_on_system = os.path.join(save_to, folder_name)

        if download:
            self.download()
        else:
            print("Verifying existing files.")
            for (recording, md5_hash) in self.recordings:
                if not check_integrity(
                    os.path.join(self.location_on_system, recording), md5_hash
                ):
                    raise RuntimeError(
                        "Dataset file not found or corrupted."
                        + " You can use download=True to download it"
                    )

    def __getitem__(self, index):
        file_path = os.path.join(self.location_on_system, self.recordings[index][0])
        topics = importRosbag(filePathOrName=file_path, log="ERROR")
        events = topics["/dvs/events"]
        events = np.stack((events["ts"], events["x"], events["y"], events["pol"])).T
        imu = topics["/dvs/imu"]
        images = topics["/dvs/image_raw"]
#         images["frames"] = np.stack(images["frames"])

        if self.transform is not None:
            events = self.transform(
                events, self.sensor_size, self.ordering, images=images
            )
        return events, imu, images

    def __len__(self):
        return len(self.recordings)

    def download(self):
        for (recording, md5_hash) in self.recordings:
            download_url(
                self.base_url + recording,
                self.location_on_system,
                filename=recording,
                md5=md5_hash,
            )



import numpy as np
import tonic.transforms as transforms


def plot_event_grid(events, ordering, axis_array=(1, 3), plot_frame_number=False):
    """Plot events accumulated in a voxel grid for visual inspection.

    Args:
        events: event Tensor of shape [num_events, num_event_channels]
        ordering: ordering of the event tuple inside of events,
                    for example 'xytp'.
        axis_array: dimensions of plotting grid. The larger the grid,
                    the more fine-grained the events will be sliced in time.
        plot_frame_number: optional index of frame when plotting

    Returns:
        None
    """
    try:
        from matplotlib import animation, rc
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError(
            "Please install the matplotlib package to plot events. This is an optional dependency."
        )

    events = events.squeeze()
    events = np.array(events)
    transform = transforms.Compose(
        [transforms.ToVoxelGrid(num_time_bins=np.product(axis_array))]
    )
    x_index = ordering.find("x")
    y_index = ordering.find("y")
    sensor_size_x = int(events[:, x_index].max() + 1)
    sensor_size_y = int(events[:, y_index].max() + 1)
    sensor_size = (sensor_size_x, sensor_size_y)

    volume = transform(events, sensor_size=sensor_size, ordering=ordering)
    fig, axes_array = plt.subplots(*axis_array)

    if 1 in axis_array:
        for i in range(np.product(axis_array)):
            axes_array[i].imshow(volume[i, :, :])
            axes_array[i].axis("off")
            if plot_frame_number:
                axes_array[i].title.set_text(str(i))
    else:
        for i in range(axis_array[0]):
            for j in range(axis_array[1]):
                axes_array[i, j].imshow(volume[i * axis_array[1] + j, :, :])
                axes_array[i, j].axis("off")
                if plot_frame_number:
                    axes_array[i, j].title.set_text(str(i * axis_array[1] + j))
    plt.tight_layout()
    plt.show()


def pad_tensors(batch):
    """This is a custom collate function for a pytorch dataloader to load multiple
    event recordings at once. It's intended to be used in combination with sparse tensors.
    All tensor sizes are extended to the largest one in the batch, i.e. the longest recording.

    Example:
        >>> dataloader = torch.utils.data.DataLoader(dataset,
        >>>                                          batch_size=10,
        >>>                                          collate_fn=tonic.utils.pad_tensors,
        >>>                                          shuffle=True)

    """
    import torch

    if not isinstance(batch[0][0], torch.Tensor):
        print(
            "tonic.utils.pad_tensors expects a PyTorch Tensor of events. Please use ToSparseTensor or similar transform to convert the events."
        )
        return None, None
    max_length = max([sample.size()[0] for sample, target in batch])

    samples_output = []
    targets_output = []
    for sample, target in batch:
        sample.sparse_resize_(
            (max_length, *sample.size()[1:]), sample.sparse_dim(), sample.dense_dim()
        )
        samples_output.append(sample)
        targets_output.append(target)
    return torch.stack(samples_output), targets_output



import numpy as np


def get_indices_values(events, sensor_size, ordering, merge_polarities):
    """Sparse Tensor PyTorch representation. See https://pytorch.org/docs/stable/sparse.html for details
    about sparse tensors. A sparse tensor will use the events as indices in the order (tpxy) and values
    of 1 for each index, which signify a spike. The shape of the tensor will be (TCWH).

    Parameters:
        merge_polarities (bool): flag that decides whether to combine polarities into a single channel
                                or split them into separate channels. If True, the number of channels for
                                indices is 1, otherwise it's the number of different polarities. Regardless 
                                of this flag, all values assigned to indices will be 1, which signify a spike.

    Returns:
        sparse tensor in TxCxWxH format, where T is timesteps, C is the number of channels for each polarity,
        and W and H are always the size of the sensor.
    """ 
    assert "x" and "t" and "p" in ordering
    x_index = ordering.find("x")
    t_index = ordering.find("t")
    p_index = ordering.find("p")

    if len(events.shape) != 2:
        raise RuntimeError(
            "Will only convert to sparse tensor from array of shape (N,E)."
        )

    # in any case, all the values in the sparse tensor will be 1, signifying a spike
    values = np.ones(events.shape[0])

    # prevents polarities used as indices that are not 0
    if len(np.unique(events[:, p_index])) == 1: merge_polarities = True
        
    if merge_polarities:  # the indices need to start at 0
        events[:, p_index] = 0
        n_channels = 1
    else:  # avoid any negative indices
        events[events[:, p_index] == -1, p_index] = 0
        n_channels = len(np.unique(events[:, p_index]))

    max_time = int(max(events[:, t_index]) + 1)
    
    if "y" in ordering:   
        y_index = ordering.find("y")
        indices = events[:, [t_index, p_index, x_index, y_index]]
    else:
        indices = events[:, [t_index, p_index, x_index]]

    return indices, values, max_time, n_channels

def to_sparse_tensor_pytorch(events, sensor_size, ordering, merge_polarities):
    try:
        import torch
    except ImportError:
        raise ImportError('The sparse tensor transform needs PyTorch installed. Please install a stable version ' +
                          'of PyTorch or alternatively install Tonic with optional PyTorch dependencies.')
    indices, values, max_time, n_channels = get_indices_values(events, sensor_size, ordering, merge_polarities)
    indices = torch.LongTensor(indices).T
    values = torch.LongTensor(values)
    return torch.sparse.FloatTensor(
        indices, values, torch.Size([max_time, n_channels, *sensor_size])
    )

def to_sparse_tensor_tensorflow(events, sensor_size, ordering, merge_polarities):
    indices, values, max_time, n_channels = get_indices_values(events, sensor_size, ordering, merge_polarities)
    try:
        import tensorflow
    except ImportError:
        raise ImportError('The sparse tensor transform needs PyTorch installed. Please install a stable version ' +
                          'of PyTorch or alternatively install Tonic with optional PyTorch dependencies.')
    return tensorflow.sparse.SparseTensor(
        indices, values, torch.Size([max_time, n_channels, *sensor_size])
    )



import os
import numpy as np
import h5py
from .dataset import Dataset
from .download_utils import (
    check_integrity,
    download_and_extract_archive,
    extract_archive,
)


class HSD(Dataset):
    """Heidelberg Spiking Dataset <https://arxiv.org/abs/1910.07407> contains the Spiking Heidelberg Dataset (SHD) 
    and the Spiking Speech Commands dataset (SSC)."""

    base_url = "https://zenkelab.org/datasets/"
    sensor_size = (700,)
    ordering = "txp"

    def __getitem__(self, index):
        file = h5py.File(os.path.join(self.location_on_system, self.filename), "r")
        # adding artificial polarity of 1
        events = np.vstack((file["spikes/times"][index], file["spikes/units"][index], np.ones(file["spikes/times"][index].shape[0]))).T
        # convert to microseconds
        events[:,0] *= 1e6
        target = file["labels"][index].astype(int)
        if self.transform is not None:
            events = self.transform(events, self.sensor_size, self.ordering)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return events, target

    def __len__(self):
        file = h5py.File(os.path.join(self.location_on_system, self.filename), "r")
        return len(file["labels"])

    def download(self):
        download_and_extract_archive(
            self.url, self.location_on_system, filename=self.zipfile, md5=self.file_md5
        )


class SHD(HSD):
    """Spiking Heidelberg Dataset. One of two Heidelberg Spiking Datasets <https://arxiv.org/abs/1910.07407>.
    Events have (txp) ordering.

    Parameters:
        save_to (string): Location to save files to on disk. Will put files in an 'hsd' subfolder.
        train (bool): If True, uses training subset, otherwise testing subset.
        download (bool): Choose to download data or verify existing files. If True and a file with the same 
                    name and correct hash is already in the directory, download is automatically skipped.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        
    Returns:
        A dataset object that can be indexed or iterated over. One sample returns a tuple of (events, targets).
    """

    test_zip = "shd_test.h5.zip"
    train_zip = "shd_train.h5.zip"
    test_md5 = "1503a5064faa34311c398fb0a1ed0a6f"
    train_md5 = "f3252aeb598ac776c1b526422d90eecb"

    def __init__(
        self, save_to, train=True, download=True, transform=None, target_transform=None
    ):
        super(HSD, self).__init__(
            save_to, transform=transform, target_transform=target_transform
        )
        self.location_on_system = os.path.join(save_to, "hsd")

        if train:
            self.url = self.base_url + self.train_zip
            self.zipfile = self.train_zip
            self.file_md5 = self.train_md5
        else:
            self.url = self.base_url + self.test_zip
            self.zipfile = self.test_zip
            self.file_md5 = self.test_md5
        self.filename = self.zipfile[:-4]

        if download:
            self.download()

        if not check_integrity(
            os.path.join(self.location_on_system, self.zipfile), self.file_md5
        ):
            raise RuntimeError(
                "Dataset not found or corrupted."
                + " You can use download=True to download it"
            )

        file = h5py.File(os.path.join(self.location_on_system, self.filename), "r")
        self.classes = file["extra/keys"]


class SSC(HSD):
    """Spiking Speech Commands dataset. One of two Heidelberg Spiking Datasets <https://arxiv.org/abs/1910.07407>.
    Events have (txp) ordering.

    Parameters:
        save_to (string): Location to save files to on disk. Will put files in an 'hsd' subfolder.
        split (string): One of 'train', 'test' or 'valid'.
        download (bool): Choose to download data or verify existing files. If True and a file with the same 
                    name and correct hash is already in the directory, download is automatically skipped.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        
    Returns:
        A dataset object that can be indexed or iterated over. One sample returns a tuple of (events, targets).
    """

    test_zip = "ssc_test.h5.zip"
    train_zip = "ssc_train.h5.zip"
    valid_zip = "ssc_valid.h5.zip"
    test_md5 = "a35ff1e9cffdd02a20eb850c17c37748"
    train_md5 = "d102be95e7144fcc0553d1f45ba94170"
    valid_md5 = "b4eee3516a4a90dd0c71a6ac23a8ae43"

    def __init__(
        self,
        save_to,
        split="train",
        download=True,
        transform=None,
        target_transform=None,
    ):
        super(HSD, self).__init__(
            save_to, transform=transform, target_transform=target_transform
        )
        self.location_on_system = os.path.join(save_to, "hsd")

        if split == "train":
            self.url = self.base_url + self.train_zip
            self.zipfile = self.train_zip
            self.file_md5 = self.train_md5
        elif split == "test":
            self.url = self.base_url + self.test_zip
            self.zipfile = self.test_zip
            self.file_md5 = self.test_md5
        elif split == "valid":
            self.url = self.base_url + self.valid_zip
            self.zipfile = self.valid_zip
            self.file_md5 = self.valid_md5
        self.filename = self.zipfile[:-4]

        if download:
            self.download()

        if not check_integrity(
            os.path.join(self.location_on_system, self.zipfile), self.file_md5
        ):
            raise RuntimeError(
                "Dataset not found or corrupted."
                + " You can use download=True to download it"
            )

        file = h5py.File(os.path.join(self.location_on_system, self.filename), "r")
        self.classes = file["extra/keys"]



import os
import numpy as np
from .dataset import Dataset
from .download_utils import (
    check_integrity,
    download_and_extract_archive,
    extract_archive,
)


class NCALTECH101(Dataset):
    """N-CALTECH101 dataset <https://www.garrickorchard.com/datasets/n-caltech101>. Events have (xytp) ordering.

    Parameters:
        save_to (string): Location to save files to on disk.
        download (bool): Choose to download data or verify existing files. If True and a file with the same 
                    name and correct hash is already in the directory, download is automatically skipped.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        
    Returns:
        A dataset object that can be indexed or iterated over. One sample returns a tuple of (events, targets).
    """

    url = "https://www.dropbox.com/sh/iuv7o3h2gv6g4vd/AADYPdhIBK7g_fPCLKmG6aVpa?dl=1"
    archive_filename = "N-Caltech101-archive.zip"
    archive_md5 = "989af2c704103341d616b748b5daa70c"
    file_md5 = "66201824eabb0239c7ab992480b50ba3"
    filename = "Caltech101.zip"
    folder_name = "Caltech101"

    sensor_size = None  # all recordings are of different size
    ordering = "xytp"

    def __init__(self, save_to, download=True, transform=None, target_transform=None):
        super(NCALTECH101, self).__init__(
            save_to, transform=transform, target_transform=target_transform
        )

        self.location_on_system = os.path.join(save_to, "ncaltech-101/")
        self.samples = []
        self.targets = []
        self.x_index = self.ordering.find("x")
        self.y_index = self.ordering.find("y")

        if download:
            self.download()

        if not check_integrity(
            os.path.join(self.location_on_system, self.filename), self.file_md5
        ):
            raise RuntimeError(
                "Dataset not found or corrupted."
                + " You can use download=True to download it."
            )

        file_path = os.path.join(self.location_on_system, self.folder_name)
        for path, dirs, files in os.walk(file_path):
            dirs.sort()
            for file in files:
                if file.endswith("bin"):
                    self.samples.append(path + "/" + file)
                    label_number = os.path.basename(path)
                    self.targets.append(label_number)

    def __getitem__(self, index):
        events = self._read_dataset_file(self.samples[index])
        target = self.targets[index]
        events[:, self.x_index] -= events[:, self.x_index].min()
        events[:, self.y_index] -= events[:, self.y_index].min()
        if self.transform is not None:
            sensor_size_x = int(events[:, self.x_index].max() + 1)
            sensor_size_y = int(events[:, self.y_index].max() + 1)
            sensor_size = (sensor_size_x, sensor_size_y)
            events = self.transform(events, sensor_size, self.ordering)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return events, target

    def __len__(self):
        return len(self.samples)

    def download(self):
        download_and_extract_archive(
            self.url,
            self.location_on_system,
            filename=self.archive_filename,
            md5=self.archive_md5,
        )
        extract_archive(os.path.join(self.location_on_system, self.filename))

    def _read_dataset_file(self, filename):
        f = open(filename, "rb")
        raw_data = np.fromfile(f, dtype=np.uint8)
        f.close()
        raw_data = np.uint32(raw_data)

        all_y = raw_data[1::5]
        all_x = raw_data[0::5]
        all_p = (raw_data[2::5] & 128) >> 7  # bit 7
        all_ts = (
            ((raw_data[2::5] & 127) << 16) | (raw_data[3::5] << 8) | (raw_data[4::5])
        )

        # Process time stamp overflow events
        time_increment = 2 ** 13
        overflow_indices = np.where(all_y == 240)[0]
        for overflow_index in overflow_indices:
            all_ts[overflow_index:] += time_increment

        # Everything else is a proper td spike
        td_indices = np.where(all_y != 240)[0]

        td = np.empty([td_indices.size, 4], dtype=np.int32)
        td[:, 0] = all_x[td_indices]
        td[:, 1] = all_y[td_indices]
        td[:, 2] = all_ts[td_indices]
        td[:, 3] = all_p[td_indices]

        return td



import os
import numpy as np
from importRosbag.importRosbag import importRosbag
from .dataset import Dataset
from .download_utils import check_integrity, download_url


class MVSEC(Dataset):
    """The Multi Vehicle Stereo Event Camera Dataset <https://daniilidis-group.github.io/mvsec/>. Events have (xytp) ordering.
    
    Parameters:
        save_to (string): Location to save files to on disk.
        scene (string): Choose one of 4 scenes: outdoor_night, outdoor_day, indoor_flying, motorcycle
        download (bool): Choose to download data or verify existing files. If True and a file with the same name is in the directory, 
                        it will be verified and re-download is automatically skipped. If False, existing files will be
                        be verified. If you already have the data on your system, make sure to place it in a subfolder
                        'MVSEC/{scene}', where {scene} is one of the available strings (see parameter above).
        transform (callable, optional): A callable of transforms to apply to events and / or images for both left and right cameras.
        
    Returns:
        A dataset object that can be indexed or iterated over. One sample returns a mix of data and ground truth in a tuple of 
        (events_left, events_right, imu_left, imu_right, images_left, images_right, depth_rect_left, depth_rect_right, pose).
    """

    resources = {
        "outdoor_night": [
            ['outdoor_night1_data.bag', '534bea503649eeb2801316704d6ab041'],
            ['outdoor_night1_gt.bag', '7e169e4048307e01f1a7ba5931ca7d4d'],
            ['outdoor_night2_data.bag', '371ff73324ba94ecb368b4c220dc8e54'],
            ['outdoor_night2_gt.bag', 'd235d28f1c93203a1d7738f8e7a67ca3'],
            ['outdoor_night3_data.bag', 'fc40889a48de7b28e6e2506125b229ac'],
            ['outdoor_night3_gt.bag', '2bbdffc94f8dd54f71486fcecfc82fe6'],
        ],
        "motorcycle": [
            ['motorcycle_data.bag', 'ae1d929563c63c4d15e0e3d3412d41c4'],
            ['motorcycle_gt.bag', 'bebf1cd58837abd1ca625ba219b47388'], 
        ],
        "indoor_flying": [
            ['indoor_flying1_data.bag', 'fd01f35eb52a754e8195d478ed2a00a2'],
            ['indoor_flying1_gt.bag', '277b97ad46f9dba3896f651ca47297aa'],
            ['indoor_flying2_data.bag', '015ac75086b248167e9602b72485e0eb'],
            ['indoor_flying2_gt.bag', '3c7ba64cd7bede77b1809cc151e54ed2'],
            ['indoor_flying3_data.bag', 'fda3408d3c72b2d7445540e5bdbd6396'],            
            ['indoor_flying3_gt.bag', 'cae5648c84ec09d0316a8d4805dee62e'],
            ['indoor_flying4_data.bag', '30ba3a744dcd4408fc4102e88c636acc'],
            ['indoor_flying4_gt.bag', 'f050c886fb34f3890fcf85680e267a21'],
        ],
        "outdoor_day": [
            ['outdoor_day1_data.bag', '7438c34b71d08ff38f52cef68834e9be'],
            ['outdoor_day1_gt.bag', '36ec3dcd0a222c4c2102641a2dc91ff0'],
            ['outdoor_day2_data.bag', '536d20bc59720b995df49f925f96b74d'],
            ['outdoor_day2_gt.bag', '69fb399411d7098b3e2cf3850f593e7b'],
        ],
    }

    base_url = "http://visiondata.cis.upenn.edu/mvsec/"
    sensor_size = (346,260)
    ordering = "xytp"

    def __init__(
        self, save_to, scene, download=True, transform=None, target_transform=None,
    ):
        super(MVSEC, self).__init__(
            save_to, transform=transform, target_transform=target_transform
        )
        self.location_on_system = os.path.join(save_to, "MVSEC")
        self.scene = scene
        if not scene in self.resources.keys():
            raise RuntimeError("Scene {} is not available or in the wrong format. Select one of: indoor_flying, outdoor_day, outdoor_night, motorcycle. ".format(scene))

        if download:
            self.download()
        else:
            print("Checking folder {}".format(os.path.join(self.location_on_system, self.scene)))
            for (filename, md5_hash) in self.resources[self.scene]:
                print("Checking integrity of file {}".format(filename))
                if not check_integrity(
                    os.path.join(self.location_on_system, self.scene, filename),
                    md5_hash,
                ):
                    raise RuntimeError(
                        "File not found or corrupted."
                        + " You can use download=True to download it"
                    )

    def __getitem__(self, index):
        # decode data file
        filename = os.path.join(self.location_on_system, self.scene, self.resources[self.scene][index*2][0])
        topics = importRosbag(filename, log="ERROR")
        events_left = topics["/davis/left/events"]
        events_left = np.stack((events_left["x"], events_left["y"], events_left["ts"]-events_left["ts"][0], events_left["pol"])).T
        events_right = topics["/davis/right/events"]
        events_right = np.stack((events_right["x"], events_right["y"], events_right["ts"]-events_right["ts"][0], events_right["pol"])).T
        imu_left = topics["/davis/left/imu"]
        imu_right = topics["/davis/right/imu"]
        images_left = topics["/davis/left/image_raw"]
        images_left = np.stack(images_left["frames"])
        images_right = topics["/davis/right/image_raw"]
        images_right = np.stack(images_right["frames"])
        
        # decode ground truth file
        filename = os.path.join(self.location_on_system, self.scene, self.resources[self.scene][index*2+1][0])
        topics = importRosbag(filename, log="ERROR")
        depth_left = topics["/davis/left/depth_image_raw"]
        depth_left = np.stack(depth_left["frames"])
        depth_right = topics["/davis/right/depth_image_raw"]
        depth_right = np.stack(depth_right["frames"])
        depth_rect_left = topics["/davis/left/depth_image_rect"]
        depth_rect_left = np.stack(depth_rect_left["frames"])
        depth_rect_right = topics["/davis/right/depth_image_rect"]
        depth_rect_right = np.stack(depth_rect_right["frames"])
        pose = topics["/davis/left/pose"]

        if self.transform is not None:
            events_left = self.transform(events_left, self.sensor_size, self.ordering, images=images_left)
        if self.transform is not None:
            events_right = self.transform(events_right, self.sensor_size, self.ordering, images=images_right)
        return events_left, events_right, imu_left, imu_right, images_left, images_right, depth_rect_left, depth_rect_right, pose

    def __len__(self):
        return len(self.resources[self.scene]) // 2  # divided by two because of data and ground truth file per recording 

    def download(self):
        for (filename, md5_hash) in self.resources[self.scene]:
            url = os.path.join(self.base_url, self.scene, filename)
            print("Downloading {}...".format(filename))
            download_url(url, os.path.join(self.location_on_system, self.scene), filename=filename, md5=md5_hash)



import numpy as np


def time_jitter_numpy(
    events,
    ordering,
    std=1,
    integer_jitter=False,
    clip_negative=False,
    sort_timestamps=False,
):
    """Changes timestamp for each event by drawing samples from a
    Gaussian distribution with the following properties:

        mean = [t]
        std = std

    Will clip negative timestamps by default.

    Parameters:
        events: ndarray of shape [num_events, num_event_channels]
        ordering: ordering of the event tuple inside of events. This function requires 't'
                  to be in the ordering
        std: change the standard deviation of the time jitter
        integer_jitter: will round the jitter that is added to timestamps
        clip_negative: drops events that have negative timestamps
        sort_timestamps: sort the events by timestamps

    Returns:
        temporally jittered set of events.
    """

    assert "t" in ordering

    t_index = ordering.find("t")
    shifts = np.random.normal(0, std, len(events))

    if integer_jitter:
        shifts = shifts.round()

    times = events[:, t_index]

    if np.issubdtype(events.dtype, np.integer):
        times += shifts.astype(np.int)
    else:
        times += shifts

    if clip_negative:
        events = np.delete(events, (np.where(times < 0)), axis=0)

    if sort_timestamps:
        events = events[np.argsort(events[:, t_index]), :]

    return events



def mask_hot_pixel(events, sensor_size, coordinates, ordering):
    """Drops events for certain pixel locations, to suppress pixels that constantly fire (e.g. due to faulty hardware).

    Parameters:
        events: ndarray of shape [num_events, num_event_channels]
        coordinates: list of (x,y) coordinates for which all events will be deleted.
        sensor_size: size of the sensor that was used [W,H]
        ordering: ordering of the event tuple inside of events. This function requires 'x' and
                  'y' to be in the ordering

    Returns:
        masked set of events.
    """

    assert "x" and "y" in ordering

    x_index = ordering.find("x")
    y_index = ordering.find("y")

    for (x, y) in coordinates:
        xs = events[:, x_index] == x
        ys = events[:, y_index] == y
        events = events[~(xs & ys)]

    return events



from .asl_dvs import ASLDVS
from .dvsgesture import DVSGesture
from .davisdataset import DAVISDATA
from .dataset import Dataset
from .hsd import SHD, SSC
from .mvsec import MVSEC
from .nmnist import NMNIST
from .ncaltech101 import NCALTECH101
from .ncars import NCARS
from .ntidigits import NTIDIGITS
from .pokerdvs import POKERDVS
from .navgesture import NavGesture
from .s_mnist import SMNIST
from .visual_place_recognition import VPR

__all__ = [
    "ASLDVS",
    "DVSGesture",
    "DAVISDATA",
    "MVSEC",
    "NMNIST",
    "NCALTECH101",
    "NCARS",
    "NTIDIGITS",
    "POKERDVS",
    "NavGesture",
    "SHD",
    "SSC",
    "SMNIST",
    "VPR",
]



import os
import numpy as np
import scipy.io as scio
from .dataset import Dataset
from .download_utils import (
    check_integrity,
    download_and_extract_archive,
    extract_archive,
)


class ASLDVS(Dataset):
    """ASL-DVS dataset <https://github.com/PIX2NVS/NVS2Graph>. Events have (txyp) ordering.

    Parameters:
        save_to (string): Location to save files to on disk.
        download (bool): Choose to download data or verify existing files. If True and a file with the same 
                    name and correct hash is already in the directory, download is automatically skipped.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
    
    Returns:
        A dataset object that can be indexed or iterated over. One sample returns a tuple of (events, targets).
    """

    url = "https://www.dropbox.com/sh/ibq0jsicatn7l6r/AACNrNELV56rs1YInMWUs9CAa?dl=1"
    filename = "ASLDVS.zip"
    file_md5 = "20f1dbf961f9a45179f6e489e93c8f2c"

    classes = [chr(letter) for letter in range(97, 123)]  # generate alphabet
    int_classes = dict(zip(classes, range(len(classes))))
    sensor_size = (240, 180)
    ordering = "txyp"

    def __init__(self, save_to, download=True, transform=None, target_transform=None):
        super(ASLDVS, self).__init__(
            save_to, transform=transform, target_transform=target_transform
        )

        self.location_on_system = save_to
        self.folder_name = "asl-dvs"
        self.data = []
        self.targets = []

        self.location_on_system = os.path.join(
            self.location_on_system, self.folder_name
        )
        if not os.path.exists(self.location_on_system):
            os.mkdir(self.location_on_system)

        if download:
            self.download()

        if not check_integrity(
            os.path.join(self.location_on_system, self.filename), self.file_md5
        ):
            raise RuntimeError(
                "Dataset not found or corrupted."
                + " You can use download=True to download it"
            )

        self.samples = []
        for path, dirs, files in os.walk(self.location_on_system):
            dirs.sort()
            files.sort()
            for file in files:
                if file.endswith("mat"):
                    self.samples.append(path + "/" + file)
                    self.targets.append(self.int_classes[path[-1]])

    def __getitem__(self, index):
        events, target = scio.loadmat(self.samples[index]), self.targets[index]
        events = (
            np.array(
                [
                    events["ts"],
                    events["x"],
                    self.sensor_size[1] - events["y"],
                    events["pol"],
                ]
            )
            .squeeze()
            .T.astype(float)
        )
        if self.transform is not None:
            events = self.transform(events, self.sensor_size, self.ordering)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return events, target

    def __len__(self):
        return len(self.samples)

    def download(self):
        download_and_extract_archive(
            self.url, self.location_on_system, filename=self.filename, md5=self.file_md5
        )
        for path, dirs, files in os.walk(self.location_on_system):
            dirs.sort()
            for file in files:
                if file.startswith("Yin") and file.endswith("zip"):
                    extract_archive(os.path.join(self.location_on_system, file))



from . import functional


class Compose:
    """Composes several transforms together.

    Parameters:
        transforms (list of ``Transform`` objects): list of transform(s) to compose. Even when using a single transform, the Compose wrapper is necessary.

    Example:
        >>> transforms.Compose([
        >>>     transforms.Denoise(),
        >>>     transforms.ToTensor(),
        >>> ])
    """

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        for t in self.transforms:
            events, images = t(
                events=events,
                images=images,
                sensor_size=sensor_size,
                ordering=ordering,
                multi_image=multi_image,
            )
        if multi_image and images.all() != None:
            return events, images
        else:
            return events

    def __repr__(self):
        format_string = self.__class__.__name__ + "("
        for t in self.transforms:
            format_string += "\n"
            format_string += "    {0}".format(t)
        format_string += "\n)"
        return format_string


class Crop:
    """Crops the sensor size to a smaller sensor.
    Removes events outsize of the target sensor and maps.

    x' = x - new_sensor_start_x

    y' = y - new_sensor_start_y

    Parameters:
        target_size: size of the sensor that was used [W',H']
    """

    def __init__(self, target_size):
        self.target_size = target_size

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        return functional.crop_numpy(
            events=events,
            sensor_size=sensor_size,
            ordering=ordering,
            images=images,
            multi_image=multi_image,
            target_size=self.target_size,
        )


class Denoise:
    """Drops events that are 'not sufficiently connected to other events in the recording.'
    In practise that means that an event is dropped if no other event occured within a spatial neighbourhood
    of 1 pixel and a temporal neighbourhood of filter_time time units. Useful to filter noisy recordings
    where events occur isolated in time.

    Parameters:
        filter_time (float): maximum temporal distance to next event, otherwise dropped.
                    Lower values will mean higher constraints, therefore less events.
    """

    def __init__(self, filter_time=10000):
        self.filter_time = filter_time

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        events = functional.denoise_numpy(
            events=events,
            sensor_size=sensor_size,
            ordering=ordering,
            filter_time=self.filter_time,
        )
        return events, images


class DropEvents:
    """Randomly drops events with drop_probability.

    Parameters:
        drop_probability (float): probability of dropping out event.
        random_drop_probability (bool): randomize the dropout probability
                                 between 0 and drop_probability.
    """

    def __init__(self, drop_probability=0.5, random_drop_probability=False):
        self.drop_probability = drop_probability
        self.random_drop_probability = random_drop_probability

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        events = functional.drop_events_numpy(
            events, self.drop_probability, self.random_drop_probability
        )
        return events, images


class FlipLR:
    """Flips events and images in x. Pixels map as:

        x' = width - x

    Parameters:
        flip_probability (float): probability of performing the flip
    """

    def __init__(self, flip_probability=0.5):
        self.flip_probability_lr = flip_probability

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        return functional.flip_lr_numpy(
            events=events,
            sensor_size=sensor_size,
            ordering=ordering,
            images=images,
            multi_image=multi_image,
            flip_probability=self.flip_probability_lr,
        )


class FlipPolarity:
    """Flips polarity of individual events with flip_probability.
    Changes polarities 1 to -1 and polarities [-1, 0] to 1

    Parameters:
        flip_probability (float): probability of flipping individual event polarities
    """

    def __init__(self, flip_probability=0.5):
        self.flip_probability_pol = flip_probability

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        events = functional.flip_polarity_numpy(
            events=events, ordering=ordering, flip_probability=self.flip_probability_pol
        )
        return events, images


class FlipUD:
    """
    Flips events and images in y. Pixels map as:

        y' = height - y

    Parameters:
        flip_probability (float): probability of performing the flip
    """

    def __init__(self, flip_probability=0.5):
        self.flip_probability_ud = flip_probability

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        return functional.flip_ud_numpy(
            events=events,
            sensor_size=sensor_size,
            ordering=ordering,
            images=images,
            multi_image=multi_image,
            flip_probability=self.flip_probability_ud,
        )


class MaskHotPixel:
    """Drops events for certain pixel locations, to suppress pixels that constantly fire (e.g. due to faulty hardware).

    Parameters:
        coordinates: list of (x,y) coordinates for which all events will be deleted.
    """

    def __init__(self, coordinates):
        self.coordinates = coordinates

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        events = functional.mask_hot_pixel(
            events=events,
            sensor_size=sensor_size,
            ordering=ordering,
            coordinates=self.coordinates,
        )
        return events, images


class RefractoryPeriod:
    """Sets a refractory period for each pixel, during which events will be
    ignored/discarded. We keep events if:

        .. math::
            t_n - t_{n-1} > t_{refrac}

    Parameters:
        refractory_period (float): refractory period for each pixel in time unit
    """

    def __init__(self, refractory_period):
        self.refractory_period = refractory_period

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        events = functional.refractory_period_numpy(
            events, sensor_size, ordering, self.refractory_period
        )
        return events, images


class SpatialJitter:
    """Changes position for each pixel by drawing samples from a multivariate
    Gaussian distribution with the following properties:

        mean = [x,y]
        covariance matrix = [[variance_x, sigma_x_y],[sigma_x_y, variance_y]]

    Jittered events that lie outside the focal plane will be dropped if clip_outliers is True.

    Parameters:
        variance_x (float): squared sigma value for the distribution in the x direction
        variance_y (float): squared sigma value for the distribution in the y direction
        sigma_x_y (float): changes skewness of distribution, only change if you want shifts along diagonal axis.
        integer_coordinates (bool): when True, shifted x and y values will be integer coordinates
        clip_outliers (bool): when True, events that have been jittered outside the focal plane will be dropped.
    """

    def __init__(
        self,
        variance_x=1,
        variance_y=1,
        sigma_x_y=0,
        integer_coordinates=True,
        clip_outliers=True,
    ):
        self.variance_x = variance_x
        self.variance_y = variance_y
        self.sigma_x_y = sigma_x_y
        self.integer_coordinates = integer_coordinates
        self.clip_outliers = clip_outliers

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        events = functional.spatial_jitter_numpy(
            events,
            sensor_size,
            ordering,
            self.variance_x,
            self.variance_y,
            self.sigma_x_y,
            self.integer_coordinates,
            self.clip_outliers,
        )
        return events, images


class SpatioTemporalTransform:
    """Transform all events spatial and temporal locations based on
    given spatial transform matrix and temporal transform vector.

    Parameters:
        spatial_transform: 3x3 matrix which can be used to perform rigid (translation and rotation),
                           non-rigid (scaling and shearing), and non-affine transformations. Generic to user input.
        temporal_transform: scale time between events and offset temporal location based on 2 member vector.
                            Used as arguments to time_skew method.
        roll: boolean input to determine if transformed events will be translated across sensor boundaries (True).
              Otherwise, events will be clipped at sensor boundaries.
    """

    def __init__(self, spatial_transform, temporal_transform, roll=False):
        self.spatial_transform = spatial_transform
        self.temporal_transform = temporal_transform
        self.roll = roll

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        events = functional.st_transform(
            events,
            sensor_size,
            ordering,
            self.spatial_transform,
            self.temporal_transform,
            self.roll,
        )
        return events, images


class TimeJitter:
    """Changes timestamp for each event by drawing samples from a
    Gaussian distribution with the following properties:

        mean = [t]
        std = std

    Will clip negative timestamps by default.

    Parameters:
        std (float): change the standard deviation of the time jitter
        integer_jitter (bool): will round the jitter that is added to timestamps
        clip_negative (bool): drops events that have negative timestamps
        sort_timestamps (bool): sort the events by timestamps
    """

    def __init__(
        self, std=1, integer_jitter=False, clip_negative=False, sort_timestamps=False
    ):
        self.std = std
        self.integer_jitter = integer_jitter
        self.clip_negative = clip_negative
        self.sort_timestamps = sort_timestamps

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        events = functional.time_jitter_numpy(
            events,
            ordering,
            self.std,
            self.integer_jitter,
            self.clip_negative,
            self.sort_timestamps,
        )
        return events, images


class TimeReversal:
    """Temporal flip is defined as:

        .. math::
           t_i' = max(t) - t_i

           p_i' = -1 * p_i

    Parameters:
        flip_probability (float): probability of performing the flip
    """

    def __init__(self, flip_probability=0.5):
        self.flip_probability_t = flip_probability

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        return functional.time_reversal_numpy(
            events=events,
            sensor_size=sensor_size,
            ordering=ordering,
            images=images,
            multi_image=multi_image,
            flip_probability=self.flip_probability_t,
        )


class TimeSkew:
    """Skew all event timestamps according to a linear transform,
    potentially sampled from a distribution of acceptable functions.

    Parameters:
        coefficient (float): a real-valued multiplier applied to the timestamps of the events.
                     E.g. a coefficient of 2.0 will double the effective delay between any
                     pair of events.
        offset (int): value by which the timestamps will be shifted after multiplication by
                the coefficient. Negative offsets are permissible but may result in
                in an exception if timestamps are shifted below 0.
    """

    def __init__(self, coefficient, offset=0):
        self.coefficient = coefficient
        self.offset = offset

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        events = functional.time_skew_numpy(
            events, ordering, self.coefficient, self.offset
        )
        return events, images


class UniformNoise:
    """
    Introduces a fixed number of noise depending on sensor size and noise
    density factor, uniformly distributed across the focal plane and in time.

    Parameters:
        scaling_factor_to_micro_sec: this is a scaling factor to get to micro
                                     seconds from the time resolution used in the event stream,
                                     as the noise time resolution is fixed to 1 micro second.
        noise_density: A noise density of 1 will mean one noise event for every
                       pixel of the sensor size for every micro second.
    """

    def __init__(self, scaling_factor_to_micro_sec=1, noise_density=1e-8):
        self.scaling_factor_to_micro_sec = scaling_factor_to_micro_sec
        self.noise_density = noise_density

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        surfaces = functional.uniform_noise_numpy(
            events,
            sensor_size,
            ordering,
            self.scaling_factor_to_micro_sec,
            self.noise_density,
        )
        return events, images


class ToAveragedTimesurface:
    """Representation that creates averaged timesurfaces for each event for one recording. Taken from the paper
    Sironi et al. 2018, HATS: Histograms of averaged time surfaces for robust event-based object classification
    https://openaccess.thecvf.com/content_cvpr_2018/papers/Sironi_HATS_Histograms_of_CVPR_2018_paper.pdf

    Parameters:
        cell_size (int): size of each square in the grid
        surface_size (int): has to be odd
        time_window (float): how far back to look for past events for the time averaging
        tau (float): time constant to decay events around occuring event with.
        decay (str): can be either 'lin' or 'exp', corresponding to linear or exponential decay.
        merge_polarities (bool): flag that tells whether polarities should be taken into account separately or not.
    """

    def __init__(
        self,
        cell_size=10,
        surface_size=7,
        temporal_window=5e5,
        tau=5e3,
        decay="lin",
        merge_polarities=False,
    ):
        assert surface_size % 2 == 1
        self.cell_size = cell_size
        self.surface_size = surface_size
        self.temporal_window = temporal_window
        self.tau = tau
        self.decay = decay
        self.merge_polarities = merge_polarities

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        surfaces = functional.to_averaged_timesurface(
            events,
            sensor_size,
            ordering,
            self.cell_size,
            self.surface_size,
            self.temporal_window,
            self.tau,
            self.decay,
            self.merge_polarities,
        )
        return surfaces, images


class ToFrame:
    """Accumulate events to frames by slicing along constant time (time_window),
    constant number of events (spike_count) or constant number of frames (n_time_bins / n_event_bins).
    You can set one of the first 4 parameters to choose the slicing method. Depending on which method you choose,
    overlap will assume different functionality, whether that might be temporal overlap, number of events
    or fraction of a bin. As a rule of thumb, here are some considerations if you are unsure which slicing
    method to choose:

    * If your recordings are of roughly the same length, a safe option is to set time_window. Bare in mind
      that the number of events can vary greatly from slice to slice, but will give you some consistency when
      training RNNs or other algorithms that have time steps.

    * If your recordings have roughly the same amount of activity / number of events and you are more interested
      in the spatial composition, then setting spike_count will give you frames that are visually more consistent.

    * The previous time_window and spike_count methods will likely result in a different amount of frames for each
      recording. If your training method benefits from consistent number of frames across a dataset (for easier
      batching for example), or you want a parameter that is easier to set than the exact window length or number
      of events per slice, consider fixing the number of frames by setting n_time_bins or n_event_bins. The two
      methods slightly differ with respect to how the slices are distributed across the recording. You can define
      an overlap between 0 and 1 to provide some robustness.

    Parameters:
        time_window (float): time window length for one frame. Use the same time unit as timestamps in the event recordings.
                             Good if you want temporal consistency in your training, bad if you need some visual consistency
                             for every frame if the recording's activity is not consistent.
        spike_count (int): number of events per frame. Good for training CNNs which do not care about temporal consistency.
        n_time_bins (int): fixed number of frames, sliced along time axis. Good for generating a pre-determined number of
                           frames which might help with batching.
        n_event_bins (int): fixed number of frames, sliced along number of events in the recording. Good for generating a
                            pre-determined number of frames which might help with batching.
        overlap (float): overlap between frames defined either in time units, number of events or number of bins between 0 and 1.
        include_incomplete (bool): if True, includes overhang slice when time_window or spike_count is specified.
                                   Not valid for bin_count methods.
        merge_polarities (bool): if True, merge polarity channels to a single channel.
    """

    def __init__(
        self,
        time_window=None,
        spike_count=None,
        n_time_bins=None,
        n_event_bins=None,
        overlap=0.0,
        include_incomplete=False,
        merge_polarities=False,
    ):
        self.time_window = time_window
        self.spike_count = spike_count
        self.n_time_bins = n_time_bins
        self.n_event_bins = n_event_bins
        self.overlap = overlap
        self.include_incomplete = include_incomplete
        self.merge_polarities = merge_polarities

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        frames = functional.to_frame_numpy(
            events=events,
            sensor_size=sensor_size,
            ordering=ordering,
            time_window=self.time_window,
            spike_count=self.spike_count,
            n_time_bins=self.n_time_bins,
            n_event_bins=self.n_event_bins,
            overlap=self.overlap,
            include_incomplete=self.include_incomplete,
            merge_polarities=self.merge_polarities,
        )
        return frames, images


class ToSparseTensor:
    """Turn event array (N,E) into sparse Tensor (B,T,W,H) if E is 4 (mostly event camera recordings),
    otherwise into sparse tensor (B,T,W) for mostly audio recordings."""

    def __init__(self, type="pytorch", merge_polarities=False):
        self.merge_polarities = merge_polarities
        self.type = type

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        if self.type == "pytorch" or "pt" or "torch":
            tensor = functional.to_sparse_tensor_pytorch(
                events=events,
                sensor_size=sensor_size,
                ordering=ordering,
                merge_polarities=self.merge_polarities,
            )
        elif self.type == "tensorflow" or "tf":
            tensor = functional.to_sparse_tensor_tensorflow(
                events=events,
                sensor_size=sensor_size,
                ordering=ordering,
                merge_polarities=self.merge_polarities,
            )
        else:
            raise NotImplementedError
        return tensor, images


class ToTimesurface:
    """Representation that creates timesurfaces for each event in the recording. Modeled after the paper
    Lagorce et al. 2016, Hots: a hierarchy of event-based time-surfaces for pattern recognition
    https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=7508476

    Parameters:
        surface_dimensions (int, int): width does not have to be equal to height, however both numbers have to be odd.
        tau (float): time constant to decay events around occuring event with.
        decay (str): can be either 'lin' or 'exp', corresponding to linear or exponential decay.
        merge_polarities (bool): flag that tells whether polarities should be taken into account separately or not.
    """

    def __init__(
        self, surface_dimensions=(7, 7), tau=5e3, decay="lin", merge_polarities=False
    ):
        assert len(surface_dimensions) == 2
        assert surface_dimensions[0] % 2 == 1 and surface_dimensions[1] % 2 == 1
        self.surface_dimensions = surface_dimensions
        self.tau = tau
        self.decay = decay
        self.merge_polarities = merge_polarities

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        surfaces = functional.to_timesurface_numpy(
            events,
            sensor_size,
            ordering,
            self.surface_dimensions,
            self.tau,
            self.decay,
            self.merge_polarities,
        )
        return surfaces, images


class ToVoxelGrid(object):
    """Build a voxel grid with bilinear interpolation in the time domain from a set of events."""

    def __init__(self, n_time_bins):
        self.n_time_bins = n_time_bins

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        volume = functional.to_voxel_grid_numpy(
            events, sensor_size, ordering, self.n_time_bins
        )
        return volume, images


class Repeat:
    """Copies target n times. Useful to transform sample labels into sequences."""

    def __init__(self, repetitions):
        self.repetitions = repetitions

    def __call__(self, target):
        return np.tile(np.expand_dims(target, 0), [self.repetitions, 1])


class ToOneHotEncoding:
    """Transforms one or more targets into a one hot encoding scheme."""

    def __init__(self, n_classes):
        self.n_classes = n_classes

    def __call__(self, target):
        return np.eye(self.n_classes)[target]


class NumpyAsType:
    def __init__(self, cast_to):
        self.cast_to = cast_to

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        events = events.astype(self.cast_to)

        if images is not None:
            images = images.astype(self.cast_to)

        return events, images



import os
import numpy as np
import h5py
from .dataset import Dataset
from .download_utils import check_integrity, download_url


class NTIDIGITS(Dataset):
    """N-TIDIGITS <https://docs.google.com/document/d/1Uxe7GsKKXcy6SlDUX4hoJVAC0-UkH-8kr5UXp0Ndi1M/edit>.
    Events have (txp) ordering.

    Parameters:
        save_to (string): Location to save files to on disk.
        train (bool): If True, uses training subset, otherwise testing subset.
        download (bool): Choose to download data or verify existing files. If True and a file with the same 
                    name and correct hash is already in the directory, download is automatically skipped.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        
    Returns:
        A dataset object that can be indexed or iterated over. One sample returns a tuple of (events, targets).
    """

    #     url = "https://www.dropbox.com/s/vfwwrhlyzkax4a2/n-tidigits.hdf5?dl=1"
    url = "https://www.neuromorphic-vision.com/public/downloads/n-tidigits.hdf5"
    file_md5 = "360a2d11e5429555c9197381cf6b58e0"
    filename = "n-tidigits.hdf5"

    sensor_size = (64,)
    ordering = "txp"

    def __init__(
        self, save_to, train=True, download=True, transform=None, target_transform=None,
    ):
        super(NTIDIGITS, self).__init__(
            save_to, transform=transform, target_transform=target_transform
        )
        self.train = train
        self.location_on_system = save_to

        if download:
            self.download()

        if not check_integrity(
            os.path.join(self.location_on_system, self.filename), self.file_md5
        ):
            raise RuntimeError(
                "Dataset not found or corrupted."
                + " You can use download=True to download it"
            )

    def __getitem__(self, index):
        file = h5py.File(os.path.join(self.location_on_system, self.filename), "r")
        if self.train:
            target = bytes.decode(file["train_labels"][index])
            timestamps = np.array(file["train_timestamps/" + target])
            addresses = np.array(file["train_addresses/" + target])
        else:
            target = bytes.decode(file["test_labels"][index])
            timestamps = np.array(file["test_timestamps/" + target])
            addresses = np.array(file["test_addresses/" + target])
        # convert timestamps to microseconds
        timestamps *= 10e5
        events = np.vstack((timestamps, addresses, np.ones(timestamps.shape[0]))).T
        if self.transform is not None:
            events = self.transform(events, self.sensor_size, self.ordering)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return events, target

    def __len__(self):
        file = h5py.File(os.path.join(self.location_on_system, self.filename), "r")
        if self.train:
            return len(file["train_labels"])
        else:
            return len(file["test_labels"])

    def download(self):
        download_url(
            self.url, self.location_on_system, filename=self.filename, md5=self.file_md5
        )



import os
import loris
import numpy
from .dataset import Dataset
from .download_utils import (
    check_integrity,
    download_and_extract_archive,
    extract_archive,
)


class NavGesture(Dataset):
    """NavGesture dataset <https://www.neuromorphic-vision.com/public/downloads/navgesture/>. Events have (txyp) ordering.

    Parameters:
        save_to (string): Location to save files to on disk.
        walk_subset (bool): Choose either NavGesture-sit (default) or NavGesture-walk dataset. No train/test split provided.
        download (bool): Choose to download data or verify existing files. If True and a file with the same 
                    name and correct hash is already in the directory, download is automatically skipped.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        
    Returns:
        A dataset object that can be indexed or iterated over. One sample returns a tuple of (events, targets).
    """

    base_url = "https://www.neuromorphic-vision.com/public/downloads/navgesture/"
    sit_filename = "navgesture-sit.zip"
    walk_filename = "navgesture-walk.zip"
    sit_url = base_url + sit_filename
    walk_url = base_url + walk_filename
    sit_md5 = "1571753ace4d9e0946e6503313712c22"
    walk_md5 = "5d305266f13005401959e819abe206f0"

    classes = ["swipe down", "swipe up", "swipe left", "swipe right", "select", "home"]
    class_codes = ["do", "up", "le", "ri", "se", "ho"]
    int_classes = dict(zip(class_codes, range(len(class_codes))))
    sensor_size = (304, 240)
    ordering = "txyp"

    def __init__(
        self,
        save_to,
        walk_subset=False,
        download=True,
        transform=None,
        target_transform=None,
    ):
        super(NavGesture, self).__init__(
            save_to, transform=transform, target_transform=target_transform
        )

        self.walk_subset = walk_subset
        self.location_on_system = save_to
        self.data = []
        self.targets = []

        if walk_subset:
            self.url = self.walk_url
            self.file_md5 = self.walk_md5
            self.filename = self.walk_filename
            self.folder_name = "navgesture-walk"
        else:
            self.url = self.sit_url
            self.file_md5 = self.sit_md5
            self.filename = self.sit_filename
            self.folder_name = "navgesture-sit"

        self.location_on_system = os.path.join(
            self.location_on_system, self.folder_name
        )
        if not os.path.exists(self.location_on_system):
            os.mkdir(self.location_on_system)

        if download:
            self.download()

        if not check_integrity(
            os.path.join(self.location_on_system, self.filename), self.file_md5
        ):
            raise RuntimeError(
                "Dataset not found or corrupted."
                + " You can use download=True to download it"
            )

        self.samples = []
        for path, dirs, files in os.walk(self.location_on_system):
            dirs.sort()
            files.sort()
            for file in files:
                if file.endswith("dat"):
                    self.samples.append(path + "/" + file)
                    self.targets.append(self.int_classes[file[7:9]])

    def __getitem__(self, index):
        events, target = loris.read_file(self.samples[index]), self.targets[index]
        events = numpy.lib.recfunctions.structured_to_unstructured(
            events["events"], dtype=float
        )
        if self.transform is not None:
            events = self.transform(events, self.sensor_size, self.ordering)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return events, target

    def __len__(self):
        return len(self.samples)

    def download(self):
        download_and_extract_archive(
            self.url, self.location_on_system, filename=self.filename, md5=self.file_md5
        )
        for path, dirs, files in os.walk(self.location_on_system):
            dirs.sort()
            for file in files:
                if file.startswith("user") and file.endswith("zip"):
                    extract_archive(os.path.join(self.location_on_system, file))



import numpy as np
import math
from .utils import slice_by_time, slice_by_spike_count, slice_by_time_bins, slice_by_event_bins


# adapted in parts from https://gitlab.com/synsense/aermanager/-/blob/master/aermanager/dataset_generator.py
def to_frame_numpy(
    events, sensor_size, ordering, time_window=None, spike_count=None, n_time_bins=None, n_event_bins=None, overlap=0., include_incomplete=False, merge_polarities=False
):
    """Accumulate events to frames by slicing along constant time (time_window), 
    constant number of events (spike_count) or constant number of frames (n_time_bins / n_event_bins).

    Parameters:
        events: ndarray of shape [num_events, num_event_channels]
        sensor_size: size of the sensor that was used [W,H]
        ordering: ordering of the event tuple inside of events.
        time_window (None): window length in us.
        spike_count (None): number of events per frame.
        n_time_bins (None): fixed number of frames, sliced along time axis.
        n_event_bins (None): fixed number of frames, sliced along number of events in the recording.
        overlap (0.): overlap between frames defined either in time in us, number of events or number of bins.
        include_incomplete (False): if True, includes overhang slice when time_window or spike_count is specified. Not valid for bin_count methods.
        merge_polarities (False): if True, merge polarity channels to a single channel.

    Returns:
        numpy array of n rate-coded frames with channels p: (NxPxWxH)
    """
    assert "x" and "y" and "t" and "p" in ordering
    assert len(sensor_size) == 2
    if not sum(param is not None for param in [time_window, spike_count, n_time_bins, n_event_bins]) == 1:
        raise ValueError("Please assign a value to exactly one of the parameters time_window, spike_count, n_time_bins or n_event_bins.")
    x_index = ordering.find("x")
    y_index = ordering.find("y")
    t_index = ordering.find("t")
    p_index = ordering.find("p")
    n_events = len(events)

    pols = events[:, p_index]
    if merge_polarities:
        pols[pols == -1] = 1
        pols[pols == 0] = 1
    else:
        pols[pols == -1] = 0

    if time_window:
        event_slices = slice_by_time(events, ordering, time_window, overlap=overlap, include_incomplete=include_incomplete)
    elif spike_count:
        event_slices = slice_by_spike_count(events, ordering, spike_count, overlap=overlap, include_incomplete=include_incomplete)
    elif n_time_bins:
        event_slices = slice_by_time_bins(events, ordering, n_time_bins, overlap=overlap)
    elif n_event_bins:
        event_slices = slice_by_event_bins(events, ordering, n_event_bins, overlap=overlap)
    
    bins_p = len(np.unique(pols))
    bins_y, bins_x = (range(sensor_size[0] + 1), range(sensor_size[1] + 1))
    
    frames = np.empty((len(event_slices), bins_p, len(bins_y) - 1, len(bins_x) - 1), dtype=np.uint16)
    for i, event_slice in enumerate(event_slices):
        frames[i] = np.histogramdd((event_slice[:,p_index], event_slice[:,y_index], event_slice[:,x_index]),
                                   bins=(np.arange(bins_p+1), bins_y, bins_x))[0]
    return frames



import os
import numpy as np
from importRosbag.importRosbag import importRosbag
from .dataset import Dataset
from .download_utils import check_integrity, download_url


class DAVISDATA(Dataset):
    """DAVIS Event Camera dataset <http://rpg.ifi.uzh.ch/davis_data.html>. Events have (txyp) ordering.

    Parameters:
        save_to (string): Location to save files to on disk. Will save files in a sub folder 'davis_dataset'.
        recording (string): Use the name of the recording or a list thereof to load it, for example 'dynamic_6dof'
                            or ['slider_far', 'urban']. See project homepage for a list of available recordings.
                            Can use 'all' to load every recording.
        download (bool): Choose to download data or verify existing files. If True and a file with the same 
                    name and correct hash is already in the directory, download is automatically skipped.
        transform (callable, optional): A callable of transforms to apply to events and/or images.
        
    Returns:
        A dataset object that can be indexed or iterated over. One sample returns a tuple of (events, imu, images, opti_track_ground_truth).
    """

    base_url = "http://rpg.ifi.uzh.ch/datasets/davis/"
    recordings = {  # recording names and their md5 hash
        "boxes_6dof": "c919cebc25e564935abfc0a3954bf016",
        "boxes_rotation": "1ec1ef4f354ce1908ed091703c1f98d0",
        "boxes_translation": "4fd11a69015022c72b375a4d000ee8cd",
        "calibration": "6e67817c7b6d2e95ca5796e89d219bac",
        "dynamic_6dof": "16e8cb3b151da15ff1fe6019b517e695",
        "dynamic_rotation": "77a2b0a87582e10e702365a38dc0c93c",
        "dynamic_translation": "2db36fd0c2945cbace48735f0416c506",
        "hdr_boxes": "67661ed4b472189432e35e8849faf200",
        "hdr_poster": "b59e37b616a3eda81184183b12b0dce5",
        "office_spiral": "60fe934450c558aff90ff5ba6b370a05",
        "office_zigzag": "3eba8fc8f42adbfd394b630852ce1f78",
        "outdoors_running": "7db2de811cb22b71fa34abd4ab1bba6b",
        "outdoors_walking": "2ef9b03c87d3c565d30211b7dcfaabc5",
        "poster_6dof": "e42ef11f523a52f11921cdb4a0fca231",
        "poster_rotation": "abcd843a894546775e5dda3560979edf",
        "poster_translation": "bb6736c56ff38f07cbe72613b72d25ed",
        "shapes_6dof": "9b9495ace2dd82881bcc5c0620dd595f",
        "shapes_rotation": "ee436cfe74b545fa25a2534f3ef021df",
        "shapes_translation": "2a805ba32671b237e4e13c023f276be9",
        "slider_close": "9426bbb70334c7b849dd5dd38eb7f2a9",
        "slider_depth": "b38a7a373f170f4b6aeca4f36e06f71a",
        "slider_far": "0f341da6aec0fd1801129e3d3a9981fa",
        "slider_hdr_close": "34cc10dd212ca1fddd3a8584046d5d1c",
        "slider_hdr_far": "c310f4f2d62cdf7b8d1c0a49315fb253",
        "urban": "c22db0b3ecbcbba8d282b0d8c3393851",
    }

    sensor_size = (180, 240)
    ordering = "txyp"

    def __init__(
        self, save_to, recording, download=True, transform=None, target_transform=None,
    ):
        super(DAVISDATA, self).__init__(
            save_to, transform=transform, target_transform=target_transform
        )
        folder_name = "davis_dataset"
        self.location_on_system = os.path.join(save_to, folder_name)
        self.selection = (
            list(self.recordings.keys()) if recording == "all" else recording
        )
        if not isinstance(self.selection, list):
            self.selection = [self.selection]

        for recording in self.selection:
            if recording not in self.recordings:
                raise RuntimeError(
                    "Recording {} is not available or in the wrong format.".format(
                        recording
                    )
                )

        if download:
            self.download()
        else:
            for recording in self.selection:
                if not check_integrity(
                    os.path.join(self.location_on_system, recording + ".bag"),
                    self.recordings[recording],
                ):
                    raise RuntimeError(
                        "Recording not found or corrupted."
                        + " You can use download=True to download it"
                    )

    def __getitem__(self, index):
        filename = os.path.join(self.location_on_system, self.selection[index] + ".bag")
        topics = importRosbag(filename, log="ERROR")
        events = topics["/dvs/events"]
        events = np.stack((events["ts"], events["x"], events["y"], events["pol"])).T
        imu = topics["/dvs/imu"]
        images = topics["/dvs/image_raw"]
        images["frames"] = np.stack(images["frames"])
        target = topics["/optitrack/davis"]

        if self.transform is not None:
            events = self.transform(
                events, self.sensor_size, self.ordering, images=images
            )
        return events, imu, images, target

    def __len__(self):
        return len(self.selection)

    def download(self):
        for recording in self.selection:
            download_url(
                self.base_url + recording + ".bag",
                self.location_on_system,
                filename=recording + ".bag",
                md5=self.recordings[recording],
            )



import numpy as np
from .utils import is_multi_image


# Code taken from https://github.com/uzh-rpg/rpg_e2vid/blob/master/utils/inference_utils.py#L431
def to_voxel_grid_numpy(events, sensor_size, ordering, n_time_bins=10):
    """Build a voxel grid with bilinear interpolation in the time domain from a set of events.

    Parameters:
        events: ndarray of shape [num_events, num_event_channels]
        sensor_size: size of the sensor that was used [W,H].
        ordering: ordering of the event tuple inside of events. This function requires 'x', 'y',
                  't' and 'p' to be in the ordering.
        n_time_bins: number of bins in the temporal axis of the voxel grid.

    Returns:
        numpy array of n event volumes (n,w,h,t)

    """
    assert "x" in ordering and "y" in ordering
    assert "t" in ordering and "p" in ordering

    x_loc = ordering.index("x")
    y_loc = ordering.index("y")
    t_loc = ordering.index("t")
    p_loc = ordering.index("p")

    voxel_grid = np.zeros(
        (n_time_bins, sensor_size[1], sensor_size[0]), np.float32
    ).ravel()

    # normalize the event timestamps so that they lie between 0 and n_time_bins
    last_stamp = events[-1, t_loc]
    first_stamp = events[0, t_loc]
    deltaT = last_stamp - first_stamp

    if deltaT == 0:
        deltaT = 1.0

    events[:, t_loc] = (n_time_bins) * (events[:, t_loc] - first_stamp) / deltaT
    ts = events[:, t_loc]
    xs = events[:, x_loc].astype(int)
    ys = events[:, y_loc].astype(int)
    pols = events[:, p_loc]
    pols[pols == 0] = -1  # polarity should be +1 / -1

    tis = ts.astype(int)
    dts = ts - tis
    vals_left = pols * (1.0 - dts)
    vals_right = pols * dts

    valid_indices = tis < n_time_bins
    np.add.at(
        voxel_grid,
        xs[valid_indices]
        + ys[valid_indices] * sensor_size[0]
        + tis[valid_indices] * sensor_size[0] * sensor_size[1],
        vals_left[valid_indices],
    )

    valid_indices = (tis + 1) < n_time_bins
    np.add.at(
        voxel_grid,
        xs[valid_indices]
        + ys[valid_indices] * sensor_size[0]
        + (tis[valid_indices] + 1) * sensor_size[0] * sensor_size[1],
        vals_right[valid_indices],
    )

    voxel_grid = np.reshape(voxel_grid, (n_time_bins, sensor_size[1], sensor_size[0]))

    return voxel_grid



import os.path
from .download_utils import check_integrity, download_and_extract_archive


class Dataset:
    def __init__(self, save_to="./", transform=None, target_transform=None):
        self.location_on_system = save_to
        self.transform = transform
        self.target_transform = target_transform
        self.data = []
        self.targets = []

    def __repr__(self):
        return "Dataset " + self.__class__.__name__

    def download(self):
        download_and_extract_archive(
            self.url, self.location_on_system, filename=self.filename, md5=self.file_md5
        )

    def check_integrity(self):
        root = self.location_on_system
        fpath = os.path.join(root, self.filename)

        if not check_integrity(fpath, self.file_md5):
            return False
        return True